#include "enemy.h"
#include "match.h"

Enemy::Enemy(Pipeline* PIPELINE, Match* MATCH, Enemy_type* TYPE) {
	pipeline 					= PIPELINE;
	match 						= MATCH;
	type 						= TYPE;
	position 					= ho::vector2f(ho::random_float(-50, pipeline->resolution[0] + 50), -50);
	current_angle 				= ho::random_float(135, 225);
	killed 						= false;
	health 						= type->full_health;
	sprite.set_texture(type->texture);
}

Enemy::~Enemy() {}

void Enemy::update() {
	float delta = life_timer.delta();

	// Choose move angle
	if(choose_angle.elapsed() || choose_angle.get_duration() <= 0) {
		move_angle = ho::random_float(135, 225);
		choose_angle.set_duration(ho::random_float(2000, 6000));
	}

	// Turning
	float turn_amount = delta * type->turn_speed;

	// Smoothly keep inside screen
	if(position[0] - type->dimensions[0] < 0 || position[0] + type->dimensions[0] > pipeline->resolution[0]) {
		// Turn toward bottom middle of screen. turn_amount * 3 for a little boost
		current_angle = ho::tween_to(current_angle, ho::angle(position, ho::vector2f(pipeline->resolution[0] / 2, pipeline->resolution[1])), turn_amount * 3);
	}
	else {
		// Smoothly face move_angle
		current_angle = ho::tween_to(current_angle, move_angle, turn_amount);
	}

	// Move enemy
	float move_amount = delta * type->move_speed;
	position = ho::trajectory(position, current_angle, move_amount);

	// Update sprite
	sprite.set_position(position - type->sprite_offset);
	sprite.set_rotation(180 + current_angle);

	if(damaged_timer.elapsed()) {
		sprite.set_rgba(ho::vector4f().one());
	}

	// Update health bar
	ho::vector2f hb_dimensions 	= ho::vector2f(40, 3);
	ho::vector2f hb_offset		 	= ho::vector2f(hb_dimensions[0] / 2, hb_dimensions[1]);
	ho::vector2f hb_sprite_offset 	= ho::vector2f(0, type->sprite_offset[1] + 10);
	ho::vector2f hb_red_tl 		= position - hb_offset - hb_sprite_offset;
	ho::vector2f hb_red_br 		= position + hb_offset - hb_sprite_offset;
	ho::vector2f hb_green_tl 		= hb_red_tl;
	ho::vector2f hb_green_br 		= ho::vector2f(hb_green_tl[0] + (hb_dimensions[0] / type->full_health) * ho::min_max(health, 0, health), hb_red_br[1]);

	health_bar_green.box(hb_green_tl, hb_green_br, ho::vector4f().zero(), rgb_255(ho::vector4f(63, 219, 6, 1)));
	health_bar_red.box(hb_red_tl, hb_red_br, ho::vector4f().zero(), rgb_255(ho::vector4f(190, 48, 49, 1)));
	health_bar_border.box(hb_red_tl, hb_red_br, ho::vector4f(0, 0, 0, 1), ho::vector4f().zero(), 1);

	// Shooting
	if(shoot_timer.elapsed() || shoot_timer.get_duration() <= 0) {
		shoot_timer.set_duration(type->shoot_offset);

		// Create bullet
		Bullet* bullet 			= new Bullet;
		bullet->position 		= position;
		bullet->sprite.set_texture(pipeline->graphics->get_texture("enemy_bullet"));
		bullet->dimensions 		= bullet->sprite.get_dimensions();
		bullet->angle 			= angle(position, match->player_position());
		bullet->speed 			= type->bullet_speed;
		bullet->damage 			= type->bullet_damage;
		match->add_enemy_bullet(bullet);
	}

	// Remove enemy if off screen
	if(position[1] > pipeline->resolution[1] + sprite.get_dimensions()[1]) {
		killed = true;
		match->add_score(-100);
		pipeline->sound->play_sound("lost enemy");
	}
}

void Enemy::render() {
	// Render enemy ship
	pipeline->world_canvas.add(sprite);

	// Render health bar
	pipeline->overlay_canvas.add(health_bar_red);
	pipeline->overlay_canvas.add(health_bar_green);
	pipeline->overlay_canvas.add(health_bar_border);
}

std::pair<ho::vector2f, ho::vector2f> Enemy::get_collision() {
	return std::pair<ho::vector2f, ho::vector2f>(position, type->dimensions / ho::vector2f(2, 2));
}

void Enemy::damage(float AMOUNT) {
	// Color enemy
	damaged_timer.set_duration(type->damaged_offset);
	sprite.set_rgba(ho::vector4f(1, 0, 0, 0.9));

	// Deduct health
	health -= AMOUNT;

	// Check if killed
	if(health <= 0) {
		killed = true;

		// Explosion sound
		switch(ho::random_int(1, 3)) {
			case 1: pipeline->sound->play_sound("explosion1"); break;
			case 2: pipeline->sound->play_sound("explosion2"); break;
			case 3: pipeline->sound->play_sound("explosion3"); break;
		}

		// Add score
		match->add_score(type->score_killed);

		// Add explosion
		Explosion* explosion = new Explosion(pipeline);
		explosion->position = position;
		explosion->animation.sprite.set_position(explosion->position - (explosion->dimensions / ho::vector2f(2, 2)));
		match->add_explosion(explosion);
	}
	else {
		// Hit sound
		switch(ho::random_int(1, 3)) {
			case 1: pipeline->sound->play_sound("hit1"); break;
			case 2: pipeline->sound->play_sound("hit2"); break;
			case 3: pipeline->sound->play_sound("hit3"); break;
		}

		// Add score
		match->add_score(1);
	}
}

bool Enemy::dead() {
	return killed;
}
