#include <iostream>
#include "common.h"
#include "game.h"
#include "graphics.h"
#include "sound.h"
#include "time.h"
#include "input.h"

// Global variables
ho::Application* application = 0;
bool running = true;

// Functions
void run_game();

int main(int argc, char** argv) {
	// Start the window
	application = new ho::Application(ho::vector2(1280, 720), "To the moon!");
	ho::Thread game_thread(run_game);
	game_thread.realtime();

	// Wait event will block until events occur
	while(running && application->wait_event());

	// Window closed, force game to stop looping
	running = false;
	game_thread.join();

	delete application;

	std::cout << "\n[i] Reached end of main(), de-constructing...\n";

	return 0;
}

void run_game() {
	HWND window_handle = application->hwnd();
	HDC device_context = GetDC(window_handle);
	HGLRC opengl_context;

	// Setup pixel format
	{
		PIXELFORMATDESCRIPTOR pfd;
		memset(&pfd, 0, sizeof(PIXELFORMATDESCRIPTOR));

		pfd.nSize = sizeof(pfd);
		pfd.nVersion = 1;
		pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_GENERIC_ACCELERATED | PFD_DOUBLEBUFFER;
		pfd.iPixelType = PFD_TYPE_RGBA;
		pfd.iLayerType = PFD_MAIN_PLANE;

		int format = ChoosePixelFormat(device_context, &pfd);

		if(!format) {
			application->kill("Could not create pixel format");
			running = false;
		}

		if(!SetPixelFormat(device_context, format, &pfd)) {
			application->kill("Could not create pixel format");
			running = false;
		}
	}

	// Setup OpenGL
	{
		opengl_context = wglCreateContext(device_context);
		wglMakeCurrent(device_context, opengl_context);

		glDisable(GL_DEPTH_TEST);
		glDisable(GL_CULL_FACE);
		glDisable(GL_DITHER);
		glDisable(GL_LIGHTING);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glDisable(GL_ALPHA_TEST);

		glewExperimental = true;
		if(glewInit() != GLEW_OK) {
			application->kill("Could not initialise GLEW");
			running = false;
		}
	}

	// Scope so pipeline gets de-constructed before OpenGL is, otherwise ho::opengl objects cannot de-construct without crashing
	{
		Pipeline pipeline;
		double frame_delta = 0;
		ho::Timer frame_timer;

		// Setup pipeline
		{
			// Load configs
			if(!pipeline.config.load("config.txt")) {
				application->kill("Could not load \"config.txt\"");
				return;
			}

			if(!pipeline.factory.load("factory.txt")) {
				application->kill("Could not load \"factory.txt\"");
				return;
			}

			// Initialise objects
			pipeline.graphics 	= new Graphics;
			pipeline.sound 		= new Sound;
			pipeline.input 		= new Input(&pipeline);

			// Load resources
			if(!pipeline.graphics->load(pipeline.factory)) {
				application->kill("Could not load graphics");
				return;
			}

			if(!pipeline.sound->load(pipeline.factory)) {
				application->kill("Could not load sounds");
				return;
			}

			// Set pipeline from config
			pipeline.vsync 			= pipeline.config.get_string("pipeline", "vsync").to_float();
			pipeline.swap_buffers 	= (bool)pipeline.config.get_string("pipeline", "swap buffers").to_int();
			pipeline.resolution 	= ho::explosion_to_vector2(pipeline.config.get_string("pipeline", "resolution").explode(","));

			// Set canvas resolutions
			pipeline.world_canvas.set_resolution(pipeline.resolution, false);
			pipeline.overlay_canvas.set_resolution(pipeline.resolution, false);
			pipeline.hud_canvas.set_resolution(pipeline.resolution, false);

			// Update window size according to config
			application->set_window_dimensions(ho::explosion_to_vector2(pipeline.config.get_string("pipeline", "window dimensions").explode(",")));

			if(pipeline.config.section_exists("key bindings")) {
				pipeline.keys.ESC 		= pipeline.config.get_string("key bindings", "ESC").to_int();
				pipeline.keys.UP 		= pipeline.config.get_string("key bindings", "UP").to_int();
				pipeline.keys.LEFT 		= pipeline.config.get_string("key bindings", "LEFT").to_int();
				pipeline.keys.DOWN 		= pipeline.config.get_string("key bindings", "DOWN").to_int();
				pipeline.keys.RIGHT 	= pipeline.config.get_string("key bindings", "RIGHT").to_int();
				pipeline.keys.SHOOT 	= pipeline.config.get_string("key bindings", "SHOOT").to_int();
			} else {
				std::cout << "[!] Warning [key bindings] section could no be found, using defaults\n";
			}
		}

		// Initialise game after pipeline is ready
		Game game(&pipeline);

		// Main game loops
		while(running) {
			pipeline.window_dimensions = application->get_window_dimensions();
			pipeline.focused = application->is_focused();

			// Pop new input buffers (only when game in focus)
			if(pipeline.focused) {
				pipeline.input->update(application->get_key_buffer(), application->get_mouse_buffer(), application->get_cursor_client_pos());

				if(pipeline.clip_cursor)
					application->clip_cursor();
				else
					application->unclip_cursor();
			}
			else {
				pipeline.input->clear();
				application->unclip_cursor();
			}

			// If closing is triggered, close before game update and render
			if(pipeline.closing) {
				running = false;
				continue;
			}

			// Update game
			game.update();

			// Watermark
			ho::opengl::Font watermark;
			watermark.set_face(pipeline.graphics->get_font("lucon"));
			watermark.set_text("By Luka Keravcin 2014");
			watermark.set_position(ho::vector2f(10, pipeline.resolution[1] - 10));

			// Render
			{
				// Update canvases
				ho::vector2f canvas_dimensions = ho::vector2f(pipeline.window_dimensions[0], pipeline.window_dimensions[1]);
				pipeline.world_canvas.set_window_dimensions(pipeline.window_dimensions);
				pipeline.world_canvas.set_dimensions(canvas_dimensions);
				pipeline.world_canvas.set_stretch(pipeline.stretch);
				pipeline.overlay_canvas.set_window_dimensions(pipeline.window_dimensions);
				pipeline.overlay_canvas.set_dimensions(canvas_dimensions);
				pipeline.overlay_canvas.set_stretch(pipeline.stretch);
				pipeline.hud_canvas.set_window_dimensions(pipeline.window_dimensions);
				pipeline.hud_canvas.set_dimensions(canvas_dimensions);
				pipeline.hud_canvas.set_stretch(pipeline.stretch);

				bool skip = false;

				if(pipeline.vsync > 0) {
					double milliseconds_per_frame = 1000.0d / (double)pipeline.vsync;
					frame_delta += frame_timer.delta();

					if(frame_delta > milliseconds_per_frame)
						frame_delta -= milliseconds_per_frame;
					else
						skip = true;
				}

				if(!skip) {
					// Clear buffer
					glClearColor(0, 0, 0, 1);
					glClear(GL_COLOR_BUFFER_BIT);

					// Render watermark
					pipeline.hud_canvas.add(watermark);

					// Render game
					game.render();

					// Render canvases to screen
					pipeline.world_canvas.render();
					pipeline.overlay_canvas.render();
					pipeline.hud_canvas.render();

					// Flush + swap buffers
					glFlush();

					if(pipeline.swap_buffers)
						SwapBuffers(device_context);
					else {
						glFinish();
						glDrawBuffer(GL_FRONT);
					}

					// Clear canvas (otherwise previous calls will remain and keep growing the size of canvas)
					pipeline.world_canvas.clear();
					pipeline.overlay_canvas.clear();
					pipeline.hud_canvas.clear();
				}
			}
		}

		// Delete objects
		delete pipeline.graphics;
		delete pipeline.sound;
		delete pipeline.input;
	}

	wglMakeCurrent(0, 0);
	wglDeleteContext(opengl_context);
	ReleaseDC(window_handle, device_context);
}

