#include "match.h"
#include "game.h"

Match::Match(Pipeline* PIPELINE, Game* GAME) {
	pipeline 			= PIPELINE;
	game 				= GAME;
	player 				= new Player(pipeline, this);
	score 				= 0;
	score_font.set_face(pipeline->graphics->get_font("lucon"));
	high_score_font.set_face(pipeline->graphics->get_font("lucon"));
	score_font.set_position(ho::vector2f(10, 20));
	high_score_font.set_position(ho::vector2f(10, 40));

	// Enemy templates
	{
		// Enemy1
		Enemy_type* enemy1 		= new Enemy_type;
		enemy1->texture 		= pipeline->graphics->get_texture("enemy1");
		enemy1->shoot_offset 	= 7000;
		enemy1->move_speed 		= 0.04;
		enemy1->turn_speed 		= 0.02;
		enemy1->full_health 	= 6;
		enemy1->damaged_offset 	= 50;
		enemy1->dimensions 		= ho::vector2f(enemy1->texture->get_dimensions()[0], enemy1->texture->get_dimensions()[1]);
		enemy1->sprite_offset 	= enemy1->dimensions / ho::vector2f(2, 2);
		enemy1->score_killed 	= 50;
		enemy1->bullet_damage 	= 1;
		enemy1->bullet_speed 	= 0.05;
		enemy_types.push_back(enemy1);

		// Enemy2
		Enemy_type* enemy2 		= new Enemy_type;
		enemy2->texture 		= pipeline->graphics->get_texture("enemy2");
		enemy2->shoot_offset 	= 3500;
		enemy2->move_speed 		= 0.06;
		enemy2->turn_speed 		= 0.03;
		enemy2->full_health 	= 8;
		enemy2->damaged_offset 	= 50;
		enemy2->dimensions 		= ho::vector2f(enemy2->texture->get_dimensions()[0], enemy2->texture->get_dimensions()[1]);
		enemy2->sprite_offset 	= enemy2->dimensions / ho::vector2f(2, 2);
		enemy2->score_killed 	= 100;
		enemy2->bullet_damage 	= 1;
		enemy2->bullet_speed 	= 0.12;
		enemy_types.push_back(enemy2);

		// Enemy3
		Enemy_type* enemy3 		= new Enemy_type;
		enemy3->texture 		= pipeline->graphics->get_texture("enemy3");
		enemy3->shoot_offset 	= 1200;
		enemy3->move_speed 		= 0.14;
		enemy3->turn_speed 		= 0.05;
		enemy3->full_health 	= 14;
		enemy3->damaged_offset 	= 50;
		enemy3->dimensions 		= ho::vector2f(enemy3->texture->get_dimensions()[0], enemy3->texture->get_dimensions()[1]);
		enemy3->sprite_offset 	= enemy3->dimensions / ho::vector2f(2, 2);
		enemy3->score_killed 	= 150;
		enemy3->bullet_damage 	= 2;
		enemy3->bullet_speed 	= 0.25;
		enemy_types.push_back(enemy3);
	}

	// Generate star starting positions
	for(unsigned i = 0; i < MAX_STARS; ++i) {
		stars[i].position 		= ho::vector2f(ho::random_float(0, pipeline->resolution[0]), ho::random_float(-50, pipeline->resolution[1]));
		float star_size 		= ho::random_float(1, 4);
		stars[i].dimensions 	= ho::vector2f(star_size, star_size);
		float star_color 		= ho::random_float(0.5, 1);
		stars[i].color 			= ho::vector4f(star_color, star_color, star_color, 0.5);
		stars[i].speed 			= 0.07 * star_size;
	}
}

Match::~Match() {
	// Delete player
	delete player;

	// Delete enemies
	for(auto it = enemies.begin(); it != enemies.end(); ++it)
		delete *it;

	// Delete player bullets
	for(auto it = player_bullets.begin(); it != player_bullets.end(); ++it)
		delete *it;

	// Delete enemy bullets
	for(auto it = enemy_bullets.begin(); it != enemy_bullets.end(); ++it)
		delete *it;

	// Delete enemy types
	for(auto it = enemy_types.begin(); it != enemy_types.end(); ++it)
		delete *it;

	// Delete explosions
	for(auto it = explosions.begin(); it != explosions.end(); ++it)
		delete *it;

	// Delete particles
	for(auto it = particles.begin(); it != particles.end(); ++it)
		delete *it;
}

void Match::update() {
	pipeline->clip_cursor = true;
	match_delta = match_timer.delta();

	update_background();
	spawn_enemies();
	player->update();
	update_enemies();
	update_bullets();
	check_collisions();
	update_explosions();
	update_particles();

	// Score
	score_font.set_text("SCORE " + ho::String(score));
	int high_score = game->get_highscore();

	if(high_score > 0) {
		high_score_font.set_text("HIGH SCORE " + ho::String(high_score));
	}

	// Set highscore when player dies
	if(player->dead())
		game->set_highscore(score);

	// Update player levels based on score
	if(score < 500)
		player->set_level(1);
	else if(score < 2500)
		player->set_level(2);
	else if(score < 5500)
		player->set_level(3);
	else if(score < 15000)
		player->set_level(4);
	else
		player->set_level(5);
}

void Match::render() {
	// Background
	pipeline->world_canvas.add(background);

	// Stars
	for(unsigned i = 0; i < MAX_STARS; ++i)
		pipeline->world_canvas.add(stars[i].primitive);

	// Player bullets
	for(auto it = player_bullets.begin(); it != player_bullets.end(); ++it)
		pipeline->world_canvas.add((*it)->sprite);

	// Enemy bullets
	for(auto it = enemy_bullets.begin(); it != enemy_bullets.end(); ++it)
		pipeline->world_canvas.add((*it)->sprite);

	// Player
	player->render();

	// Enemies
	for(auto it = enemies.begin(); it != enemies.end(); ++it)
		(*it)->render();

	// Explosions
	for(auto it = explosions.begin(); it != explosions.end(); ++it)
		pipeline->world_canvas.add((*it)->animation.sprite);

	// Particles
	for(auto it = particles.begin(); it != particles.end(); ++it)
		pipeline->world_canvas.add((*it)->primitive);

	// Score
	pipeline->hud_canvas.add(score_font);
	pipeline->hud_canvas.add(high_score_font);

	// DEBUG
	/*{
		for(auto e = enemies.begin(); e != enemies.end(); ++e) {
			Enemy* enemy = *e;
			auto e_pair = enemy->get_collision();
			ho::vector2f e_pos = e_pair.first;
			ho::vector2f e_rad = e_pair.second;
			ho::opengl::Primitive2d e_prim;
			e_prim.ellipse(e_pos, e_rad, 16, ho::vector4f(0, 1, 0, 1), ho::vector4f().zero(), 1);
			pipeline->overlay_canvas.add(e_prim);

			for(auto pb = player_bullets.begin(); pb != player_bullets.end(); ++pb) {
				Bullet* bullet = *pb;
				auto b_pair = bullet->get_collision();
				ho::vector2f b_pos = b_pair.first;
				ho::vector2f b_rad = b_pair.second;
				ho::opengl::Primitive2d prim;
				float langle = angle(b_pos, e_pos);
				ho::vector2f offset = trajectory_position(b_pos, langle, radius(b_rad, langle));
				ho::vector4f points = ho::vector4f(b_pos[0], b_pos[1], offset[0], offset[1]);
				prim.line(points, ho::vector4f(1, 0, 0, 1), 1);
				pipeline->overlay_canvas.add(prim);
			}
		}
	}*/
}

void Match::add_player_bullet(Bullet* BULLET) {
	player_bullets.push_back(BULLET);
}

void Match::add_enemy_bullet(Bullet* BULLET) {
	enemy_bullets.push_back(BULLET);
}

void Match::add_score(int AMOUNT) {
	score += AMOUNT;
}

void Match::add_explosion(Explosion* EXPLOSION) {
	explosions.push_back(EXPLOSION);

	// Generate some particles
	for(unsigned i = 0; i < 40; ++i) {
		Particle* p 			= new Particle;
		p->position 			= EXPLOSION->position;
		float particle_size 	= ho::random_float(1, 4);
		p->dimensions 			= ho::vector2f(particle_size, particle_size);
		p->angle 				= ho::random_float(0, 360);
		p->speed 				= 0.07 * particle_size;
		p->color 				= ho::vector4f(1, ho::random_float(0.2, 1), 0, 1);

		p->life.set_duration(ho::random_float(1500, 4000));
		particles.push_back(p);
	}
}

ho::vector2f Match::player_position() {
	auto player_collision = player->get_collision();

	return player_collision.first;
}

void Match::update_background() {
	ho::vector4f background_color = rgb_255(ho::vector4f(45, 45, 75, 1));
	background.box(pipeline->screen_bounds, ho::vector4f().zero(), background_color, 0);

	for(unsigned i = 0; i < MAX_STARS; ++i) {
		Particle* star = &stars[i];

		// Update position
		star->position = ho::trajectory(star->position, 180, match_delta * star->speed);

		// Regenerate if off screen
		if(star->position[1] > pipeline->resolution[1] + star->dimensions[1]) {
			star->position 		= ho::vector2f(ho::random_float(0, pipeline->resolution[0]), -10);
			float star_size 	= ho::random_float(1, 5);
			star->dimensions 	= ho::vector2f(star_size, star_size);
			float star_color 	= ho::random_float(0.5, 1);
			star->color 		= ho::vector4f(star_color, star_color, star_color, 0.5);
			star->speed 		= 0.07 * star_size;
		}

		// Update primitive
		star->primitive.box(star->position, star->position + star->dimensions, ho::vector4f().zero(), star->color);
	}
}

void Match::update_bullets() {
	// Player bullets
	for(auto it = player_bullets.begin(); it != player_bullets.end();) {
		Bullet* bullet = *it;

		// Move bullet
		bullet->position = ho::trajectory(bullet->position, bullet->angle, match_delta * bullet->speed);

		// Update sprite
		ho::vector2f bullet_offset = bullet->sprite.get_dimensions() / ho::vector2f(2, 2);
		bullet->sprite.set_position(bullet->position - bullet_offset);
		bullet->sprite.set_rotation(bullet->angle);

		// Remove if off screen (player bullets only go up)
		if(bullet->position[1] < 0 - bullet->sprite.get_dimensions()[1]) {
			delete bullet;
			it = player_bullets.erase(it);
		}
		else
			++it;
	}

	// Enemy bullets
	for(auto it = enemy_bullets.begin(); it != enemy_bullets.end();) {
		Bullet* bullet = *it;

		// Move bullet
		bullet->position = ho::trajectory(bullet->position, bullet->angle, match_delta * bullet->speed);

		// Update sprite
		ho::vector2f bullet_offset = bullet->sprite.get_dimensions() / ho::vector2f(2, 2);
		bullet->sprite.set_position(bullet->position - bullet_offset);
		bullet->sprite.set_rotation(bullet->angle);

		// Remove if off screen (player bullets only go up)
		if(!point_inside_box(bullet->position, ho::vector2f(-bullet->dimensions[0], -bullet->dimensions[1]), ho::vector2f(pipeline->resolution[0] + bullet->dimensions[0], pipeline->resolution[1] + bullet->dimensions[1]))) {
			delete bullet;
			it = enemy_bullets.erase(it);
		}
		else
			++it;
	}
}

void Match::check_collisions() {
	auto player_collision 			= player->get_collision();
	ho::vector2f player_position 	= player_collision.first;
	ho::vector2f player_radius 	= player_collision.second;

	for(auto e = enemies.begin(); e != enemies.end(); ++e) {
		Enemy* enemy 					= *e;
		auto enemy_collision 			= enemy->get_collision();
		ho::vector2f enemy_position 	= enemy_collision.first;
		ho::vector2f enemy_radius 		= enemy_collision.second;

		// Player bullet hit enemies
		for(auto pb = player_bullets.begin(); pb != player_bullets.end();) {
			Bullet* bullet 					= *pb;
			auto bullet_collision 			= bullet->get_collision();
			ho::vector2f bullet_position 	= bullet_collision.first;
			ho::vector2f bullet_radius 	= bullet_collision.second;

			if(ho::radial_intercept(bullet_position, enemy_position, bullet_radius, enemy_radius)) {
				enemy->damage(bullet->damage);
				delete bullet;
				pb = player_bullets.erase(pb);
			}
			else
				++pb;
		}

		if(!player->dead() && ho::radial_intercept(player_position, enemy_position, player_radius, enemy_radius)) {
			enemy->damage(10);
			player->damage(10);
			score -= 100;
		}
	}

	// Enemy bullets hit player
	if(!player->dead()) {
		for(auto eb = enemy_bullets.begin(); eb != enemy_bullets.end();) {
			Bullet* bullet 					= *eb;
			auto bullet_collision 			= bullet->get_collision();
			ho::vector2f bullet_position 	= bullet_collision.first;
			ho::vector2f bullet_radius 	= bullet_collision.second;

			if(ho::radial_intercept(bullet_position, player_position, bullet_radius, player_radius)) {
				player->damage(bullet->damage);
				delete bullet;
				eb = enemy_bullets.erase(eb);
				score -= 10;
			}
			else
				++eb;
		}
	}
}

void Match::update_enemies() {
	for(auto it = enemies.begin(); it != enemies.end();) {
		Enemy* enemy = *it;

		if(enemy->dead()) {
			it = enemies.erase(it);
			delete enemy;
		}
		else {
			enemy->update();
			++it;
		}
	}
}

void Match::spawn_enemies() {
	if(spawner_timer.elapsed() || spawner_timer.get_duration() <= 0) {
		float time_seconds = match_timer.seconds();

		if(time_seconds < 30) {
			enemies.push_back(new Enemy(pipeline, this, enemy_types[0]));
			spawner_timer.set_duration(ho::random_float(1200, 1700));
		}
		else if(time_seconds < 90) {
			enemies.push_back(new Enemy(pipeline, this, enemy_types[0]));
			spawner_timer.set_duration(ho::random_float(750, 1400));
		}
		else if(time_seconds < 120) {
			enemies.push_back(new Enemy(pipeline, this, enemy_types[0]));
			spawner_timer.set_duration(ho::random_float(500, 800));
		}
		else if(time_seconds < 180) {
			enemies.push_back(new Enemy(pipeline, this, enemy_types[1]));
			spawner_timer.set_duration(ho::random_float(510, 750));
		}
		else {
			enemies.push_back(new Enemy(pipeline, this, enemy_types[2]));
			spawner_timer.set_duration(ho::random_float(500, 800));
		}
	}
}

void Match::update_explosions() {
	for(auto it = explosions.begin(); it != explosions.end();) {
		Explosion* explosion =  *it;
		explosion->animation.update();

		if(explosion->animation.timer.elapsed()) {
			delete explosion;
			it = explosions.erase(it);
		}
		else
			++it;
	}
}

void Match::update_particles() {
	for(auto it = particles.begin(); it != particles.end();) {
		Particle* p = *it;

		// Move particle
		p->position = ho::trajectory(p->position, p->angle, match_delta * p->speed);

		// Update color to fade
		p->color[3] = (1 - (p->life.milliseconds() / p->life.get_duration()));

		// Update primitive
		p->primitive.box(p->position, p->position + p->dimensions, ho::vector4f().zero(), p->color);

		// Remove if off screen
		if(!ho::point_inside_box(p->position, ho::vector2f(-p->dimensions[0], -p->dimensions[1]), ho::vector2f(pipeline->resolution[0] + p->dimensions[0], pipeline->resolution[1] + p->dimensions[1]))) {
			delete p;
			it = particles.erase(it);
		}
		// Remove if timer elapsed
		else if(p->life.elapsed()) {
			delete p;
			it = particles.erase(it);
		}
		else
			++it;
	}
}
