#include "player.h"
#include "match.h"

Player::Player(Pipeline* PIPELINE, Match* MATCH) {
	pipeline 					= PIPELINE;
	match  						= MATCH;
	sprite.set_texture(pipeline->graphics->get_texture("ship"));
	dimensions 					= sprite.get_dimensions();
	sprite_offset 				= dimensions / ho::vector2f(2, 2);
	position[0] 				= pipeline->resolution[0] / 2;
	position[1] 				= pipeline->resolution[1] / 2;
	fire 						= false;
	move_speed 					= 0.23;
	full_health 				= 20;
	health 						= full_health;
	killed						= false;
	damaged_offset 				= 50;
	level						= 0;
	set_level(1);
	shoot_timer.set_duration(shoot_offset);
}

Player::~Player() {}

void Player::update() {
	if(!killed) {
		handle_input();
		update_position();
		update_shooting();
	}

	ho::vector2f hb_dimensions 	= ho::vector2f(250, 10);
	ho::vector2f hb_position 		= ho::vector2f(40, pipeline->resolution[1] - 40);
	ho::vector2f hb_red_tl 		= hb_position;
	ho::vector2f hb_red_br 		= hb_position + hb_dimensions;
	ho::vector2f hb_green_tl 		= hb_red_tl;
	ho::vector2f hb_green_br 		= ho::vector2f(hb_position[0] + (hb_dimensions[0] / full_health) * ho::min_max(health, 0, health), hb_red_br[1]);

	health_bar_green.box(hb_green_tl, hb_green_br, ho::vector4f().zero(), rgb_255(ho::vector4f(63, 219, 6, 1)));
	health_bar_red.box(hb_red_tl, hb_red_br, ho::vector4f().zero(), rgb_255(ho::vector4f(190, 48, 49, 1)));
	health_bar_border.box(hb_red_tl, hb_red_br, ho::vector4f(0, 0, 0, 1), ho::vector4f().zero(), 1);
}

void Player::render() {
	if(!killed) {
		// Player ship
		pipeline->world_canvas.add(sprite);
	}

	if(damaged_timer.elapsed()) {
		sprite.set_rgba(ho::vector4f().one());
	}

	// Render health bar
	pipeline->overlay_canvas.add(health_bar_red);
	pipeline->overlay_canvas.add(health_bar_green);
	pipeline->overlay_canvas.add(health_bar_border);
}

void Player::set_level(int LEVEL) {
	if(LEVEL > level)
		pipeline->sound->play_sound("level up");

	switch(LEVEL) {
		case 1: {
			if(level != LEVEL) {
				level 			= LEVEL;
				shoot_offset 	= 210;
				weapon 			= Weapon::GUN;
			}

			break;
		}
		case 2: {
			if(level != LEVEL) {
				level 			= LEVEL;
				shoot_offset 	= 160;
				weapon 			= Weapon::GUN;
			}

			break;
		}
		case 3: {
			if(level != LEVEL) {
				level 			= LEVEL;
				shoot_offset 	= 100;
				weapon 			= Weapon::GUN;
			}

			break;
		}
		case 4: {
			if(level != LEVEL) {
				level 			= LEVEL;
				shoot_offset 	= 80;
				weapon 			= Weapon::DOUBLE_GUN;
			}

			break;
		}
		case 5: {
			if(level != LEVEL) {
				level 			= LEVEL;
				shoot_offset 	= 80;
				weapon 			= Weapon::LASER;
			}

			break;
		}
	}
}

void Player::damage(float AMOUNT) {
	// Color player
	damaged_timer.set_duration(damaged_offset);
	sprite.set_rgba(ho::vector4f(1, 0, 0, 0.9));

	// Deduct health
	health -= AMOUNT;

	// Check if killed
	if(health <= 0) {
		killed = true;

		// Explosion sound
		switch(ho::random_int(1, 3)) {
			case 1: pipeline->sound->play_sound("explosion1"); break;
			case 2: pipeline->sound->play_sound("explosion2"); break;
			case 3: pipeline->sound->play_sound("explosion3"); break;
		}

		// Add explosion
		Explosion* explosion = new Explosion(pipeline);
		explosion->position = position;
		explosion->animation.sprite.set_position(explosion->position - (explosion->dimensions / ho::vector2f(2, 2)));
		match->add_explosion(explosion);

		pipeline->sound->play_sound("player died");
	} else {
		pipeline->sound->play_sound("player hit");
	}
}

bool Player::dead() {
	return killed;
}

std::pair<ho::vector2f, ho::vector2f> Player::get_collision() {
	return std::pair<ho::vector2f, ho::vector2f>(position, dimensions / ho::vector2f(2, 2));
}

void Player::handle_input() {
	float move_amount = life_timer.delta() * move_speed;
	auto keyboard_mem = &pipeline->input->keyboard_mem[0];

	// Keyboard movement
	if(keyboard_mem[pipeline->keys.UP])
		position[1] += move_amount;
	if(keyboard_mem[pipeline->keys.LEFT])
		position[0] += move_amount;
	if(keyboard_mem[pipeline->keys.DOWN])
		position[1] -= move_amount;
	if(keyboard_mem[pipeline->keys.RIGHT])
		position[0] -= move_amount;

	// Keyboard or mouse shooting
	if(keyboard_mem[pipeline->keys.SHOOT] || pipeline->input->mouse_mem[0]) {
		if(shoot_timer.elapsed() || shoot_timer.get_duration() <= 0) {
			fire = true;
			shoot_timer.set_duration(shoot_offset);
		}
	}

	// Mouse movement
	position = pipeline->input->cursor_position;
}

void Player::update_position() {
	// Keep ship inside resolution
	ho::vector2f top_left = position - sprite_offset;
	ho::vector2f dimensions = sprite.get_dimensions();
	if(top_left[0] < 0)
		position[0] += -top_left[0];
	if(top_left[0] + dimensions[0] > pipeline->resolution[0])
		position[0] += pipeline->resolution[0] - (top_left[0] + dimensions[0]);
	if(top_left[1] < 0)
		position[1] += -top_left[1];
	if(top_left[1] + dimensions[1] > pipeline->resolution[1])
		position[1] += pipeline->resolution[1] - (top_left[1] + dimensions[1]);

	// Ship sprite position
	sprite.set_position(position - sprite_offset);
}

void Player::update_shooting() {
	if(fire) {
		switch(weapon) {
			case Weapon::GUN: {
				Bullet* bullet 			= new Bullet;
				bullet->position 		= position;
				bullet->sprite.set_texture(pipeline->graphics->get_texture("bullet"));
				bullet->dimensions 		= ho::vector2f(1, 11);
				bullet->angle 			= ho::random_float(-2.2, 2.2);
				bullet->speed 			= 0.7;
				bullet->damage 			= 1;
				match->add_player_bullet(bullet);
				pipeline->sound->play_sound("gun");

				break;
			}
			case Weapon::DOUBLE_GUN: {
				ho::vector2f offset = ho::vector2f(8, 0);

				// Bullet 1
				Bullet* bullet 			= new Bullet;
				bullet->position 		= position - offset;
				bullet->sprite.set_texture(pipeline->graphics->get_texture("bullet"));
				bullet->dimensions 		= ho::vector2f(1, 11);
				bullet->angle 			= ho::random_float(-0.8, 0.8);
				bullet->speed 			= 0.9;
				bullet->damage 			= 0.7;
				match->add_player_bullet(bullet);
				// Bullet 2
				bullet 					= new Bullet;
				bullet->position 		= position + offset;
				bullet->sprite.set_texture(pipeline->graphics->get_texture("bullet"));
				bullet->dimensions 		= ho::vector2f(1, 11);
				bullet->angle 			= ho::random_float(-0.8, 0.8);
				bullet->speed 			= 0.9;
				bullet->damage 			= 0.7;
				match->add_player_bullet(bullet);

				pipeline->sound->play_sound("gun");

				break;
			}
			case Weapon::LASER: {
				Bullet* bullet 			= new Bullet;
				bullet->position 		= position;
				bullet->sprite.set_texture(pipeline->graphics->get_texture("laser"));
				bullet->dimensions 		= ho::vector2f(1, 23);
				bullet->angle 			= ho::random_float(-0.7, 0.7);
				bullet->speed 			= 1.3;
				bullet->damage 			= 3;
				match->add_player_bullet(bullet);
				pipeline->sound->play_sound("laser");

				break;
			}
		}

		fire = false;
	}
}
